#ifndef STOER_WAGNER_MINCUT_HPP
#define STOER_WAGNER_MINCUT_HPP 1

#include <cassert>
#include <functional>
#include <map>
#include <set>
#include <utility>
#include <boost/config.hpp>
#include <boost/concept_check.hpp>
#include <boost/concept/assert.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/property_map/property_map.hpp>
#include <boost/tuple/tuple.hpp>

namespace detail {

template <class undirected_graph_, class weight_map_, class vertex_assignment_map_>
boost::tuple<BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor, BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor, BOOST_DEDUCED_TYPENAME boost::property_traits<weight_map_>::value_type> stoer_wagner_phase(const undirected_graph_& g, const weight_map_& weights, const vertex_assignment_map_& assignments);

template <class undirected_graph_, class weight_map_, class vertex_assignment_map_>
class max_adjacency_search_datum
{
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor vertex_descriptor;
	typedef BOOST_DEDUCED_TYPENAME boost::property_traits<weight_map_>::value_type weight_type;

	vertex_descriptor v;
	weight_type wA;
	
public:
	max_adjacency_search_datum(vertex_descriptor v_, weight_type wA_)
		: v(v_), wA(wA_)
	{
	}
	
	inline bool operator>(const max_adjacency_search_datum& n2) const
	{
		return (wA > n2.wA);
	}
	
	friend boost::tuple<vertex_descriptor, vertex_descriptor, weight_type> stoer_wagner_phase<undirected_graph_, weight_map_, vertex_assignment_map_>(const undirected_graph_&, const weight_map_&, const vertex_assignment_map_&);
};

/**
 * \brief Performs a phase of the Stoer–Wagner min-cut algorithm
 *
 * Performs a \em phase of the Stoer–Wagner min-cut algorithm.
 *
 * As described by Stoer & Wagner (1997), a phase is simply a maximum adjacency search
 * (also called a maximum cardinality search), which results in the minimum <em>s</em>-<em>t</em> cut
 * of the input graph. Here, the input graph is basically \p g, but some vertices
 * are virtually assigned to others as a way of viewing \p g as a graph with
 * some sets of vertices merged together.
 *
 * This implementation is a translation of pseudocode by Uri Zwick, School of
 * Computer Science, Tel Aviv University.
 *
 * \pre \p g is a simple, connected, undirected graph
 * \param[in] g the input graph
 * \param[in] weights a property map from each edge to its weight (a positive value)
 * \param[in] assignments a property map from each vertex to the vertex that it is assigned to
 * \returns a tuple (\em s, \em t, \em w) of the "<em>s</em>" and "<em>t</em>"
 * 	of the minimum <em>s</em>-<em>t</em> cut and the cut weight \em w
 * 	of the minimum <em>s</em>-<em>t</em> cut.
 * \see http://www.cs.tau.ac.il/~zwick/grad-algo-08/gmc.pdf
 *
 * \author Daniel Trebbien
 * \date 2010-06-19
 */
template <class undirected_graph_, class weight_map_, class vertex_assignment_map_>
boost::tuple<BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor, BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor, BOOST_DEDUCED_TYPENAME boost::property_traits<weight_map_>::value_type> stoer_wagner_phase(const undirected_graph_& g, const weight_map_& weights, const vertex_assignment_map_& assignments)
{
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor vertex_descriptor;
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_iterator vertex_iterator;
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::out_edge_iterator out_edge_iterator;
	typedef BOOST_DEDUCED_TYPENAME boost::property_traits<weight_map_>::value_type weight_type;
	
	typedef detail::max_adjacency_search_datum<undirected_graph_, weight_map_, vertex_assignment_map_> max_adjacency_search_datum;
	typedef std::multiset<max_adjacency_search_datum, std::greater<max_adjacency_search_datum> > max_adjacency_search_data_multiset;
	
	max_adjacency_search_data_multiset maxAdjSearchData;
	
	vertex_iterator verticesIt, verticesEnd;
	boost::tie(verticesIt, verticesEnd) = boost::vertices(g);
	for(; verticesIt != verticesEnd; ++verticesIt) {
		const vertex_descriptor v = *verticesIt;
		if (v == boost::get(assignments, v)) {
			maxAdjSearchData.insert(max_adjacency_search_datum(v, 0));
		}
	}
	
	vertex_descriptor s, t;
	weight_type w;
	while (maxAdjSearchData.size() > 0) { // while PQ \neq {} do
		BOOST_DEDUCED_TYPENAME max_adjacency_search_data_multiset::iterator masdIt = maxAdjSearchData.begin(), masdEnd = maxAdjSearchData.end();
		const vertex_descriptor u = masdIt->v;
		w = masdIt->wA;
		maxAdjSearchData.erase(masdIt++); // u = extractmax(PQ)
		
		s = t; t = u;
		while(masdIt != masdEnd) {
			out_edge_iterator outEdgesIt, outEdgesEnd;
			boost::tie(outEdgesIt, outEdgesEnd) = boost::out_edges(u, g);
			bool masdItIncremented = false;
			for(; outEdgesIt != outEdgesEnd; ++outEdgesIt) {
				const vertex_descriptor v = boost::get(assignments, boost::target(*outEdgesIt, g));
				
#ifndef NDEBUG
				const vertex_descriptor masdItV = masdIt->v; // DEBUG
#endif
				if (masdIt->v == v) { // if v \in PQ then
					assert(boost::get(weights, *outEdgesIt) > 0);
#ifndef NDEBUG
					const weight_type masdItWA = masdIt->wA; // DEBUG
					const weight_type outEdgesItWeight = boost::get(weights, *outEdgesIt); // DEBUG
#endif
					BOOST_DEDUCED_TYPENAME max_adjacency_search_data_multiset::iterator resIt = maxAdjSearchData.insert(max_adjacency_search_datum(v, masdIt->wA + boost::get(weights, *outEdgesIt)));
					assert(resIt->wA > masdIt->wA);
					maxAdjSearchData.erase(masdIt++);
					masdItIncremented = true;
					break; // assume that the graph has no parallel edges. Therefore, there isn't another edge from u to v.
				}
			}
			
			if (!masdItIncremented)
				++masdIt;
		}
	}
	
	return boost::make_tuple(s, t, w);
}

} // end `namespace detail`

/**
 * \brief Computes a min-cut of the input graph
 *
 * Computes a min-cut of the input graph using the Stoer–Wagner algorithm.
 *
 * \pre \p g is a simple, connected, undirected graph
 * \param[in] g the input graph
 * \param[in,out] weights a property map from each edge to its weight (a positive value). The
 * 	weight property values are modified by the implementation.
 * \param[out] parities a property map from each vertex to an unspecified-bool-type object for
 * 	distinguishing the two vertex sets of the min-cut
 * \param[out] assignments a property map from each vertex to a \c vertex_descriptor object. This
 * 	map serves as work space, and no particular meaning should be derived from property values
 * 	after completion of the algorithm.
 * \returns the cut weight of the min-cut
 * \see http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.114.6687&rep=rep1&type=pdf
 * \see http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.31.614&rep=rep1&type=pdf
 *
 * \author Daniel Trebbien
 * \date 2010-06-19
 */
template <class undirected_graph_, class weight_map_, class parity_map_, class vertex_assignment_map_>
BOOST_DEDUCED_TYPENAME boost::property_traits<weight_map_>::value_type stoer_wagner_mincut(const undirected_graph_& g, weight_map_& weights, parity_map_& parities, vertex_assignment_map_& assignments)
{
	BOOST_CONCEPT_ASSERT((boost::IncidenceGraphConcept<undirected_graph_>));
	BOOST_CONCEPT_ASSERT((boost::VertexListGraphConcept<undirected_graph_>));

	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_descriptor vertex_descriptor;
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertex_iterator vertex_iterator;
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::vertices_size_type vertices_size_type;
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::edge_descriptor edge_descriptor;
	typedef BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::out_edge_iterator out_edge_iterator;
	
	BOOST_CONCEPT_ASSERT((boost::ReadWritePropertyMapConcept<weight_map_, edge_descriptor>));
	typedef BOOST_DEDUCED_TYPENAME boost::property_traits<weight_map_>::value_type weight_type;
	BOOST_CONCEPT_ASSERT((boost::Convertible<BOOST_DEDUCED_TYPENAME boost::graph_traits<undirected_graph_>::directed_category, boost::undirected_tag>));
	BOOST_CONCEPT_ASSERT((boost::WritablePropertyMapConcept<parity_map_, vertex_descriptor>));
	typedef BOOST_DEDUCED_TYPENAME boost::property_traits<parity_map_>::value_type parity_type;
	BOOST_CONCEPT_ASSERT((boost::Convertible<bool, parity_type>));
	BOOST_CONCEPT_ASSERT((boost::WritablePropertyMapConcept<vertex_assignment_map_, vertex_descriptor>));
	BOOST_CONCEPT_ASSERT((boost::Convertible<vertex_descriptor, BOOST_DEDUCED_TYPENAME boost::property_traits<vertex_assignment_map_>::value_type>));
	
	vertices_size_type n = boost::num_vertices(g);
	if (n < 2)
		throw std::invalid_argument("the input graph must have at least two vertices.");
	
	// initialize `assignments` (all vertices are initially assigned to themselves)
	vertex_iterator verticesIt, verticesEnd;
	boost::tie(verticesIt, verticesEnd) = boost::vertices(g);
	for (; verticesIt != verticesEnd; ++verticesIt) {
		const vertex_descriptor v = *verticesIt;
		boost::put(assignments, v, v);
	}
	
	vertex_descriptor s, t;
	weight_type bestW;
	
	boost::tie(s, t, bestW) = detail::stoer_wagner_phase(g, weights, assignments);
	boost::tie(verticesIt, verticesEnd) = boost::vertices(g);
	for(; verticesIt != verticesEnd; ++verticesIt) {
		const vertex_descriptor v = *verticesIt;
		boost::put(parities, v, (v == t ? true : false));
	}
	boost::put(assignments, t, s);
	std::map<vertex_descriptor, edge_descriptor> sOutEdges; // from target v to edge {s, v}. We assume that there are no parallel edges in the input graph. Therefore, a `std::map` is sufficient.
	out_edge_iterator sOutEdgesIt, sOutEdgesEnd;
	boost::tie(sOutEdgesIt, sOutEdgesEnd) = boost::out_edges(s, g);
	for(; sOutEdgesIt != sOutEdgesEnd; ++sOutEdgesIt) {
		sOutEdges[boost::target(*sOutEdgesIt, g)] = *sOutEdgesIt;
	}
	out_edge_iterator tOutEdgesIt, tOutEdgesEnd;
	boost::tie(tOutEdgesIt, tOutEdgesEnd) = boost::out_edges(t, g);
	for(; tOutEdgesIt != tOutEdgesEnd; ++tOutEdgesIt) {
		BOOST_DEDUCED_TYPENAME std::map<vertex_descriptor, edge_descriptor>::iterator sOutEdge = sOutEdges.find(boost::target(*tOutEdgesIt, g));
		if (sOutEdge != sOutEdges.end()) {
			assert(boost::target(sOutEdge->second, g) == boost::target(*tOutEdgesIt, g));
#ifndef NDEBUG
			const vertex_descriptor sOutEdgeTarget = boost::target(sOutEdge->second, g); // DEBUG
			const weight_type sOutEdgeSecondWeight = boost::get(weights, sOutEdge->second), // DEBUG
				tOutEdgesItWeight = boost::get(weights, *tOutEdgesIt); // DEBUG
#endif
			boost::put(weights, sOutEdge->second, boost::get(weights, sOutEdge->second) + boost::get(weights, *tOutEdgesIt));
#ifndef NDEBUG
			const weight_type sOutEdgeSecondWeight2 = boost::get(weights, sOutEdge->second); // DEBUG
#endif
		}
	}
	--n;
	
	for (; n > 2; --n) {
		weight_type w;
		boost::tie(s, t, w) = detail::stoer_wagner_phase(g, weights, assignments);
		
		if (w < bestW) {
			boost::tie(verticesIt, verticesEnd) = boost::vertices(g);
			for(; verticesIt != verticesEnd; ++verticesIt) {
				const vertex_descriptor v = *verticesIt;
				boost::put(parities, v, (boost::get(assignments, v) == t ? true : false));
				
				if (boost::get(assignments, v) == t) // all vertices that were assigned to t are now assigned to s
					boost::put(assignments, v, s);
			}
			
			bestW = w;
		}
		else {
			boost::tie(verticesIt, verticesEnd) = boost::vertices(g);
			for(; verticesIt != verticesEnd; ++verticesIt) {
				const vertex_descriptor v = *verticesIt;
				if (boost::get(assignments, v) == t) // all vertices that were assigned to t are now assigned to s
					boost::put(assignments, v, s);
			}
		}
		
		// assign t to s
		boost::put(assignments, t, s);
		sOutEdges.clear();
		boost::tie(sOutEdgesIt, sOutEdgesEnd) = boost::out_edges(s, g);
		for(; sOutEdgesIt != sOutEdgesEnd; ++sOutEdgesIt) {
			sOutEdges[boost::target(*sOutEdgesIt, g)] = *sOutEdgesIt;
		}
		boost::tie(tOutEdgesIt, tOutEdgesEnd) = boost::out_edges(t, g);
		for(; tOutEdgesIt != tOutEdgesEnd; ++tOutEdgesIt) {
			BOOST_DEDUCED_TYPENAME std::map<vertex_descriptor, edge_descriptor>::iterator sOutEdge = sOutEdges.find(boost::target(*tOutEdgesIt, g));
			if (sOutEdge != sOutEdges.end()) {
				assert(boost::target(sOutEdge->second, g) == boost::target(*tOutEdgesIt, g));
				assert(boost::get(weights, sOutEdge->second) > 0);
				assert(boost::get(weights, *tOutEdgesIt) > 0);
				boost::put(weights, sOutEdge->second, boost::get(weights, sOutEdge->second) + boost::get(weights, *tOutEdgesIt));
			}
		}
	}

	return bestW;
}

#endif // !STOER_WAGNER_MINCUT_HPP
